#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import logging
import os
from os.path import join

from tqdm import tqdm

from moss.ems.utilities.pagination_query import EmsPaginationQueryException

logger = logging.getLogger("ems_cli_export")


class EmsCliExport:
    def __init__(self):
        # when not used in CLI
        self.cli_mode = False

    def ems_export(self, project_name, output_directory, indent, **kwargs):
        """
        Exports a complete EMS project

        Args:
            project_name (str): the name of the EMS project
            output_directory (str):  the name of the output directory

        Returns:

        Raises:
            EmsCliException
        """
        from moss.ems.scripts.moss_emscli import EmsCliException

        # Check if the project exist
        project = self.ems_service.project(project_name)  # type: ignore

        if project is None:
            logger.error(
                "The project '%s' does not exist in this instance", project_name
            )
            raise EmsCliException(
                "The project '{}' does not exist in this instance".format(project_name)
            )

        # Export the project structure and save it to file
        saved_project = self.ems_service.export_project(project_name)  # type: ignore
        if "error" in saved_project:
            raise EmsCliException(
                "An error occured in exporting the project {}".format(project_name)
            )
        self.json_to_file(
            saved_project, output_directory, "project", indent
        )  # type: ignore

        # Check if the project has variant
        objectclasses = project.objectclasses
        # Looking for the VNT Master Objectclass
        variant_master_found = None
        variant_master_found = next(
            (item for item in objectclasses if item.objectClassType == "VNTMASTER"),
            None,
        )

        if variant_master_found is not None:
            logger.info(
                "An objectclass 'VNTMASTER' found. Switching to 'variant' mode!"
            )

            vn = variant_master_found.name
            subdir = join(output_directory, vn)
            os.mkdir(subdir)

            if "master_filter" in kwargs:
                master_filter = kwargs.get("master_filter")
                logger.info("Using {} to query master".format(master_filter))
                try:
                    query_results = variant_master_found.layers[0].query(
                        where=master_filter
                    )
                except EmsPaginationQueryException:
                    logger.error(
                        "Error running query with filter '{}'".format(master_filter)
                    )
                    return
            else:
                query_results = variant_master_found.layers[0].query()

            resolved_query = list(query_results.resolve())

            featureid_name = resolved_query[0]["objectIdFieldName"]
            features = resolved_query[0]["features"]

            if self.cli_mode:
                print("Found {n} masters. Start export...".format(n=len(features)))
            with tqdm(total=len(features), disable=self.cli_mode is False) as pbar:
                for query in features:
                    featureid = query["attributes"][featureid_name]
                    subdir = os.path.join(output_directory, vn, str(featureid))
                    logger.info("Creating master feature {}".format(featureid))
                    os.mkdir(subdir)
                    self.json_to_file(query, subdir, "master", indent)
                    variants = variant_master_found.variants(featureid)
                    self.json_to_file(variants, subdir, "variants", indent)
                    for variant in variants:
                        _id = variant[featureid_name]
                        subdir = os.path.join(
                            output_directory, vn, str(featureid), str(_id)
                        )
                        logger.info("Creating variant {}".format(_id))
                        os.mkdir(subdir)
                        for objectclass in objectclasses:
                            if objectclass.has_variant:
                                subdir = os.path.join(
                                    output_directory,
                                    vn,
                                    str(featureid),
                                    str(_id),
                                    objectclass.name,
                                )
                                logger.info(
                                    "Creating objectclass {}".format(objectclass.name)
                                )
                                os.mkdir(subdir)
                                for layer in objectclass.layers:
                                    layer_data = layer.query(
                                        geometry=True, variants=[_id]
                                    )
                                    for feature_sets in layer_data.resolve():
                                        if feature_sets["count"] > 0:
                                            logger.info(
                                                "Creating layer {}".format(layer.name)
                                            )
                                            self.json_to_file(  # type: ignore
                                                feature_sets["features"],
                                                subdir,
                                                layer.name,
                                                indent,
                                            )
                    pbar.update(1)

            if self.cli_mode:
                print("Export successfully finished.")

        else:
            logger.info("No VARIANTMASTER found. There should be no variants.")
            logger.info("Starting to export data in %s", output_directory)

            for objectclass in objectclasses:
                logger.debug("Exporting objectclass %s", objectclass)
                for layer in objectclass.layers:
                    logger.debug("Exporting layer %s", layer)
                    layer_data = layer.query(returnGeometry=True)
                    for layers in layer_data.resolve():
                        if layers["count"]:
                            self.json_to_file(  # type: ignore
                                layers,
                                os.path.join(output_directory, objectclass.name),
                                "_".join([layer.name, "Layers"]),
                                indent,
                            )
        logger.info("Export completed")



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import logging
import os
import sys
from os import listdir, walk
from os.path import exists, isdir, isfile, join

from tqdm import tqdm

logger = logging.getLogger("ems_cli_import")


class EmsCliImport:
    def __init__(self):
        # when not used in CLI
        self.cli_mode = False

    def ems_import(self, project_name, input_directory):
        # type: (str, str) -> None
        """Imports a complete EMS project

        Args:
            project_name: the name of the EMS project
            input_directory:  the name of the input directory

        Returns:

        Raises:
            EmsCliException
        """
        from moss.ems.scripts.moss_emscli import EmsCliException
        from moss.ems.utilities.sortutil import ObjectClassSorter

        logger.info("Starting import process")
        logger.debug(
            "Projects in the current instance: %s",
            self.ems_service.projects,  # type: ignore
        )

        # Check if the path exist
        if not os.path.exists(input_directory):
            logger.error("The path %s does not exist")
            sys.exit(1)

        project = next(
            iter(filter(lambda p: p.name == project_name, self.ems_service.projects)),  # type: ignore
            None,
        )

        # project = None
        if project is not None:
            logger.error(
                "The project '%s' does already exist in this instance", project_name
            )
            raise EmsCliException(
                "The project '{}' does already exist in this instance".format(
                    project_name
                )
            )

        # Get the project file (the json with the complete structure)
        saved_project = self.file2json(  # type: ignore
            os.path.join(input_directory, "project.json")  # type: ignore
        )
        if "description" not in saved_project:
            raise EmsCliException("The JSON structure in project.json is wrong")
        logger.debug(
            "Importing project %s:\n\tDescription: %s",
            project_name,
            saved_project["description"],
        )
        ret_value = self.ems_service.import_project(saved_project, project_name)  # type: ignore

        if ret_value is False:
            logger.error("The project '%s' couldn't be imported", project_name)
            raise EmsCliException(
                "The project '{}' couldn't be imported".format(project_name)
            )
        project = self.ems_service.project(project_name)  # type: ignore

        # Check if the project has variant
        sorter = ObjectClassSorter(project.objectclasses)
        objectclasses = sorter.by_reference()

        # Looking for the VNT Master Objectclass
        variant_master_found = None
        logger.info(
            "Looking for a VARIANTMASTER in the project structure ... i.e. a variant"
            " project"
        )
        variant_master_found = next(
            (item for item in objectclasses if item.objectClassType == "VNTMASTER"),
            None,  # type: ignore
        )

        # Check if there is a variant tree
        if variant_master_found is not None:
            logger.info(
                "The VARIANTMASTER '{}' was found".format(variant_master_found.name)
            )
            # switching to the VNTMASTER directory
            input_directory = join(input_directory, variant_master_found.name)
            # Get all the master variants ids, i.e. the directory names
            master_variant_ids = [
                mvi
                for mvi in listdir(input_directory)
                if isdir(join(input_directory, mvi))
            ]
            logger.info(
                "{} master variants were found: {}".format(
                    len(master_variant_ids), master_variant_ids
                )
            )

            # maps an objectclass to a map of old to new ids
            objectclass_map = {}

            # maps the old master id to a new master id
            master_map = {}
            objectclass_map[variant_master_found.name] = master_map

            total_master_count = len(master_variant_ids)

            if self.cli_mode:
                print("Found {n} masters. Start import...".format(n=total_master_count))

            with tqdm(total=total_master_count, disable=self.cli_mode is False) as pbar:
                while len(master_variant_ids) > 0:
                    current_variant_id = master_variant_ids.pop(0)

                    logger.info("Creating master variant {}".format(current_variant_id))
                    master_variant_dir = join(input_directory, current_variant_id)

                    if not os.path.exists(join(master_variant_dir, "master.json")):
                        raise EmsCliException(
                            "The file 'master.json' is missing in {}".format(
                                master_variant_dir
                            )
                        )
                    logger.info("Reading the master features in 'master.json'")
                    master_features = self.file2json(  # type: ignore
                        join(master_variant_dir, "master.json")
                    )
                    preferred_field_name = ""
                    preferred_variant = ""
                    preferred_field_exisits = False
                    preferred_field = next(
                        (
                            item
                            for item in variant_master_found.fields  # type: ignore
                            if item["type"] == "emsFieldTypeVariantPref"
                        ),
                        None,
                    )
                    selfref_field = next(
                        (
                            item
                            for item in variant_master_found.fields  # type: ignore
                            if item["type"] == "emsFieldTypeObjectClass"
                            and item["objectClassName"] == variant_master_found.name
                        ),
                        None,
                    )

                    if selfref_field is not None:
                        # variant master supports self reference
                        selfref_field_name = selfref_field["name"]
                        selfref_id = master_features["attributes"][selfref_field_name]

                        # if referenced master object is not created yet, push the current
                        # master_variant_id back to stack and continue with next folder
                        if selfref_id is not None and selfref_id not in master_map:
                            master_variant_ids.append(current_variant_id)
                            continue

                    if preferred_field is not None:
                        preferred_field_exisits = True
                        preferred_field_name = preferred_field["name"]
                        preferred_variant = master_features["attributes"][
                            preferred_field_name
                        ]
                        del master_features["attributes"][preferred_field_name]

                    old_master_id = master_features["attributes"]["ID"]

                    # theese attributes need to be deleted, otherwise -> error
                    del master_features["attributes"]["GLOBALID"]
                    del master_features["attributes"]["ID"]

                    # add the new master feature
                    logger.info("Add the new master features")
                    result_dict = variant_master_found.layers[0].add_features(
                        [master_features]
                    )
                    if "code" in result_dict:
                        error_dict = result_dict
                        raise EmsCliException(
                            "Could not add master variant features:\n\tCode:"
                            " {}\n\tMessage: {}\n\tDescription: {}".format(
                                error_dict["code"],
                                error_dict["message"],
                                error_dict["description"],
                            )
                        )
                    if "error" in result_dict["addResults"][0]:
                        error_dict = result_dict["addResults"][0]["error"]
                        raise EmsCliException(
                            "Could not add master variant features:\n\tCode:"
                            " {}\n\tMessage: {}\n\tDescription: {}".format(
                                error_dict["code"],
                                error_dict["message"],
                                error_dict["description"],
                            )
                        )
                    new_master_id = result_dict["addResults"][0]["objectId"]

                    master_map[old_master_id] = new_master_id

                    # the root variant(s) of the new feature, hopefully only one
                    all_variants = variant_master_found.variants(new_master_id)
                    if len(all_variants) > 1:  # type: ignore
                        raise EmsCliException("More than one root variant was found")
                    if not exists(join(master_variant_dir, "variants.json")):
                        raise EmsCliException(
                            "The file 'variants.json' is missing in {}".format(
                                master_variant_dir
                            )
                        )
                    # read the variants for the master feature
                    logger.info(
                        "Reading the variants for the master features in"
                        " 'variants.json'"
                    )
                    exported_variants = self.file2json(  # type: ignore
                        join(master_variant_dir, "variants.json")
                    )

                    root_variant = all_variants[0]

                    featureid_name, master_results = variant_master_found.query(
                        where="ID = " + str(new_master_id)
                    )
                    master = master_results[0]

                    # maps old variant ids to new variant ids
                    variant_map = {}

                    new_variant_id = root_variant["ID"]

                    for variant in exported_variants:
                        # is it the start variant ?
                        if variant["PARENT"] != -1:
                            # we need the master_id
                            new_parent_variant_id = variant_map[variant["PARENT"]]

                            logger.info(
                                "Derive '{}' from master_id {}, variant_id {}".format(
                                    variant["NAME"],
                                    new_master_id,
                                    new_parent_variant_id,
                                )
                            )

                            result_dict = master.derive(
                                feature_id=new_master_id,
                                variant_id=new_parent_variant_id,
                                name=variant["NAME"],
                                description=variant["DESCRIPTION"],
                                status=variant["STATUS"],
                                category=variant["CATEGORY"],
                            )

                            if "error" in result_dict["deriveVariantResult"]:
                                error_dict = result_dict["deriveVariantResult"][0][
                                    "error"
                                ]
                                raise EmsCliException(
                                    "Could not derive from master variant features:"
                                    "\n\tCode: {}\n\tMessage: {}\n\tDescription: {}".format(
                                        error_dict["code"],
                                        error_dict["message"],
                                        error_dict["description"],
                                    )
                                )
                            # set the new id in the exported_variants tree to find parents
                            new_variant_id = result_dict["deriveVariantResult"][
                                "objectId"
                            ]

                            # now drain the variant, so we remove any derived objects from the new variant
                            master.drain(
                                feature_id=new_master_id,
                                variant_id=new_variant_id,
                                object_classes="*",
                            )

                        variant_map[variant["ID"]] = new_variant_id
                        variant_dir = join(master_variant_dir, str(variant["ID"]))

                        # now add the features, i.e. all the json files in the subdirectories (name=obj_class)
                        # (json file name = layer name)

                        for objectclass in objectclasses:
                            featureset_dir = join(variant_dir, objectclass.name)

                            if not isdir(featureset_dir):
                                continue
                            logger.info("Processing objectclass {}".format(objectclass))

                            # reset in each objectclass iteration
                            created_features = {}
                            failed_features = []
                            objectclass_map[objectclass.name] = created_features

                            # list of already existing feature IDs
                            old_feature_id = []

                            featureset_files = [
                                f
                                for f in listdir(featureset_dir)
                                if isfile(join(featureset_dir, f))
                            ]
                            for featureset_file in featureset_files:
                                # list of features to add
                                features_to_add = []
                                # list of features already created, which therefore must be updated
                                features_to_update = []

                                layer_name = featureset_file[
                                    0 : featureset_file.find(".json")
                                ]
                                logger.info(
                                    "Add layer {} to {}".format(
                                        layer_name, objectclass.name
                                    )
                                )
                                feature_set = self.file2json(  # type: ignore
                                    join(featureset_dir, featureset_file)
                                )

                                references = list(
                                    filter(
                                        lambda field: field["type"]
                                        == "emsFieldTypeObjectClass",
                                        objectclass.fields,
                                    )
                                )

                                for feature in feature_set:

                                    # get a reference to the old feature id
                                    old_feature_id.append(feature["attributes"]["ID"])
                                    # o_feature_id = feature["attributes"]["ID"]

                                    if old_feature_id[-1] in created_features:
                                        # already created so update the features ID
                                        new_feature_id = created_features[
                                            old_feature_id[-1]
                                        ]
                                        # n_feature_id = feature_id_map[old_feature_id]

                                        feature["attributes"]["ID"] = new_feature_id

                                        # and append it to the list of featurs to update
                                        features_to_update.append(feature)

                                    # only update the feature if it was created previously
                                    # without any errors.
                                    elif old_feature_id[-1] not in failed_features:
                                        # not yet created

                                        # The GLOBALID and ID can't be imported but most be
                                        # newly generated by the EMS
                                        del feature["attributes"]["GLOBALID"]
                                        del feature["attributes"]["ID"]
                                        for reference in references:
                                            reference_field = reference["name"]
                                            ref_obj = reference["objectClassName"]
                                            id_map = objectclass_map[ref_obj]
                                            old_id = feature["attributes"][
                                                reference_field
                                            ]
                                            new_id = id_map[old_id]
                                            feature["attributes"][
                                                reference_field
                                            ] = new_id

                                        del feature["variants"]
                                        features_to_add.append(feature)

                                # get acces to the layer by its feature_set
                                # layer = layer_from_feature_set(feature_set)
                                layer = objectclass.layer(layer_name)

                                if features_to_add:
                                    # add the not yet created features
                                    logger.info(
                                        "Add {} features to layer {} {}.".format(
                                            len(features_to_add),
                                            layer_name,
                                            featureset_dir,
                                        )
                                    )
                                    add_results = layer.add_features(
                                        features=features_to_add,
                                        variant_id=new_variant_id,
                                    )

                                    # and update the created_features map with the latest additions
                                    for idx, add_result in enumerate(
                                        add_results["addResults"]
                                    ):

                                        # since EMS garantuees that add_results are returned
                                        # in the same order as the list of added features
                                        # we can easily access the id of the created feature
                                        # in the old system

                                        added_feature = features_to_add[idx]

                                        if add_result["success"] is True:
                                            new_feature_id = add_result["objectId"]

                                            created_features[
                                                old_feature_id[idx]
                                            ] = new_feature_id
                                            # feature_id_map[o_feature_id] = new_feature_id
                                        else:
                                            logger.error(
                                                "Failed to create feature: %s",
                                                added_feature,
                                            )

                                if features_to_update:
                                    # update the already created features
                                    logger.info(
                                        "Update {} features in layer {}.".format(
                                            len(features_to_update), layer_name
                                        )
                                    )
                                    update_results = layer.update_features(
                                        features=features_to_update,
                                        variant_id=new_variant_id,
                                    )

                                    # and update the created_features map with the latest additions
                                    for idx, update_result in enumerate(
                                        update_results["updateResults"]
                                    ):
                                        if update_result["success"] is False:
                                            failed_features.append(
                                                features_to_update[idx]["attributes"][
                                                    "ID"
                                                ]
                                            )
                                            logger.error(
                                                """
                                                Failed to update feature. This means the feature was
                                                only created partialy:\n
                                                Feature-Set:\n
                                                %s\n
                                                Update Result:\n
                                                %s
                                                """,
                                                features_to_update,
                                                update_result,
                                            )

                    # Now lets restore the preferred variant ID
                    if preferred_field_exisits is True:
                        pagination_query = variant_master_found.layers[0].query(
                            where="ID={}".format(new_master_id)
                        )
                        result = list(pagination_query.resolve())[0]
                        features = result["features"]
                        feature_attributes = features[0]["attributes"]
                        try:
                            feature_attributes[preferred_field_name] = variant_map[
                                preferred_variant
                            ]
                        except KeyError:
                            logger.warning(
                                "The preferred variant %s does not exist",
                                preferred_variant,
                            )
                        result_dict = variant_master_found.layers[0].update_features(
                            features
                        )
                        if "code" in result_dict:
                            error_dict = result_dict
                            raise EmsCliException(
                                "Could not update master variant features:"
                                "\n\tCode: {}\n\tMessage: {}\n\tDescription: {}".format(
                                    error_dict["code"],
                                    error_dict["message"],
                                    error_dict["description"],
                                )
                            )
                        if "error" in result_dict["updateResults"][0]:
                            error_dict = result_dict["updateResults"][0]["error"]
                            raise EmsCliException(
                                "Could not update master variant features: "
                                "\n\tCode: {}\n\tMessage: {}\n\tDescription: {}".format(
                                    error_dict["code"],
                                    error_dict["message"],
                                    error_dict["description"],
                                )
                            )
                    pbar.update(1)

            if self.cli_mode:
                print("Import successfully finished.")
            # show the new tree
            project.variants_tree()
        else:
            objclass_name_marker = ""
            created_features = []
            failed_features = []
            for dirpath, _, filenames in walk(input_directory):
                for filename in [f for f in filenames if f.endswith("Layers.json")]:
                    # the object class name is the directory name of the JSON file
                    objclass_name = os.path.split(dirpath)[1]
                    logger.info("Processing Objectclass: %s", objclass_name)

                    # the add_feature and update_feature process is only
                    # necessary for one objectclass!
                    # Do we have a new object class?
                    if objclass_name != objclass_name_marker:
                        # new map to track which features where created
                        created_features = {}

                        # new map to track features for which creation failed
                        failed_features = []
                        objclass_name_marker = objclass_name

                    # list of features to add
                    features_to_add = []
                    # list of features already created, which therefore must be updated
                    features_to_update = []
                    # list of already existing feature IDs
                    old_feature_id = []

                    layers = self.file2json(os.path.join(dirpath, filename))  # type: ignore
                    layer_name = filename[0 : filename.find("_Layers.json")]

                    # split features into two lists: already created and not yet created
                    for feature in layers["features"]:

                        # get a reference to the old feature id
                        old_feature_id.append(feature["attributes"]["ID"])

                        if old_feature_id[-1] in created_features:
                            # already created so update the features ID
                            new_feature_id = created_features[old_feature_id[-1]]

                            feature["attributes"]["ID"] = new_feature_id

                            # and append it to the list of featurs to update
                            features_to_update.append(feature)

                        # only update the feature if it was created previously
                        # without any errors.
                        elif old_feature_id[-1] not in failed_features:
                            # not yet created

                            # The GLOBALID and ID can't be imported but most be
                            # newly generated by the EMS
                            del feature["attributes"]["GLOBALID"]
                            del feature["attributes"]["ID"]
                            features_to_add.append(feature)

                    # get acces to the layer by its feature_set
                    # layer = layer_from_feature_set(feature_set)
                    layer = project.objectclass(objclass_name).layer(layer_name)

                    if features_to_add:
                        # add the not yet created features
                        add_results = layer.add_features(features_to_add)

                        # and update the created_features map with the latest additions
                        for idx, add_result in enumerate(add_results["addResults"]):

                            # since EMS garantuees that add_results are returned
                            # in the same order as the list of added features
                            # we can easily access the id of the created feature
                            # in the old system

                            added_feature = features_to_add[idx]

                            if add_result["success"] is True:
                                new_feature_id = add_result["objectId"]

                                created_features[old_feature_id[idx]] = new_feature_id
                            else:
                                logger.error(
                                    "Failed to create feature: %sß", added_feature
                                )

                    if features_to_update:
                        # update the already created features
                        update_results = layer.update_features(features_to_update)

                        # and update the created_features map with the latest additions
                        for idx, update_result in enumerate(
                            update_results["updateResults"]
                        ):
                            if "error" in update_result:
                                failed_features.append(
                                    features_to_update[idx]["attributes"]["ID"]
                                )
                                logger.error(
                                    """
                                    Failed to update feature. This means the feature was
                                    only created partialy:\n
                                    Feature-Set:\n
                                    %s\n
                                    Update Result:\n
                                    %s
                                    """,
                                    features_to_update,
                                    update_result,
                                )
        logger.info("Import completed")



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#


import copy
import json
import logging

from .emslayer import EmsLayer
from .emsqueryresult import EmsQueryResult

logger = logging.getLogger("ems_objectclass")


class ObjectClassException(Exception):
    """
    Handles exceptions for the objectClass
    """


class ObjectClass:
    """
    The Objectclass class
    """

    def __init__(self, project, session, object_definition, count_features=True):
        self._project = project
        self._session = session
        self.name = ""
        self.variantMaster = -1
        # the whole construction with this redundant variant_master is because
        # of the name convention and the assignment:
        # "if self.variant_master" needs an assigned self.variant_master AND
        # variantMaster does not fullfill the name convention
        # self.variant_master = 0

        for key, value in object_definition.items():
            setattr(self, key, value)
        #            if key == "variantMaster":
        #                self.variant_master = value
        if count_features:
            if self.variantMaster == -1:
                self._total_features = self._count_features()
            else:
                self._total_features = 0
        else:
            logger.debug("Not counting features as requested")
            self._total_features = -1

    def _count_features(self):
        """
        Counts the features of an objectclass
        """
        try:
            first_layer = self.layers[0]
        except KeyError:
            raise ObjectClassException("There are no layers in this objectclass")

        logger.debug("First layer is %s", first_layer)
        if first_layer is not None:

            url = "/".join(
                ["rest", self._project, self.name, first_layer.name, "query"]
            )
            parameters = {"f": "json", "where": "1=1", "returnCountOnly": "true"}

            logger.debug("Count features from %s", url)

            response = self._session.get(url, params=parameters)
            try:
                response_dict = response.json()
            except KeyError:
                return None

            count = 0
            try:
                count = response_dict["count"]
            except KeyError:
                logger.exception(
                    "There is no count key in the response: %s", response_dict
                )
                error_dictionary = response_dict["error"]
                logger.exception("Message: %s", error_dictionary["message"])
                logger.exception("Description: %s", error_dictionary["description"])
                logger.exception("Code: %s", error_dictionary["code"])
                raise ObjectClassException()
            return count

        return 0

    @property
    def id_field(self):
        #  type: ()->str
        """
        Return the name of the ID field for this objectclass
        """
        query_results = self.query(where="1=1")

        if query_results:
            try:
                id_field = query_results[0]
            except IndexError:
                id_field = "ID"
            return id_field
        else:
            # TODO Probably we should raise an error
            logger.warning(
                "Something is not correct requesting ID field %s.", query_results
            )
            return "ID"

    @property
    def has_variant(self):
        """
        Check if it has variant
        """
        return self.variantMaster != -1  # pylint:disable=no-member

    @property
    def layers(self):
        """
        The Layers for this ObjectClass

        Example
        --------

        >>> layers = objectclass.layers

        Returns:
            layers: A list of layers for the objectclass
        """

        url = "/".join(["rest", self._project, self.name, "layers"])
        parameters = {"f": "json"}

        response = self._session.get(url, params=parameters)
        try:
            response_dict = response.json()
        except KeyError:
            return None

        layers = [
            EmsLayer(
                layer["name"],
                self.name,
                self._project,
                self._session,
                copy.deepcopy(layer),
            )
            for layer in response_dict["layers"]
        ]

        return layers

    def layer(self, layer_name):
        """
        Return a layer instance

        Example:
            >>> objectclass.layer("TEST_OBJECTCLASS")

        Args:
            layer_name: The name of the layer

        Returns:
            :layer: the layer instance with the given name
        """

        url = "/".join(["rest", self._project, self.name, layer_name])
        response = self._session.get(url)
        try:
            response_dict = response.json()
        except KeyError:
            return None

        if "error" in response_dict.keys():
            logger.error("The layer %s does not exist", layer_name)
            raise ObjectClassException("Layer does not exist")
        else:
            layer = EmsLayer(
                layer_name,
                self.name,
                self._project,
                self._session,
                definition=response_dict,
            )
            return layer

    @property
    def features(self):
        """
        The features of this ObjectClass

        Example:
            >>> objectclass.features()

        Args:
            -

        Returns:
            :features: the features of the objectclass
        """
        first_layer = self.layers[0]

        if first_layer is not None:
            logger.debug("Getting features for %s", first_layer.name)
            pagination_query = first_layer.query(where="1=1")
            total_results = len(pagination_query)
            if total_results:
                result = list(pagination_query.resolve())[0]
                return result["features"]
            else:
                logger.debug("No layer found for %s", first_layer)
        return []

    def query(self, where="1=1", complete=False):
        """
        Query the objectClass

        Args:
            where: The SQL Expression to query the objectClass

        Returns:
            :list:
                - first element: the field name of the object id (default ID)
                - second element: a list of feature attributes
        """

        first_layer = self.layers[0]

        if first_layer:

            url = "/".join(
                ["rest", self._project, self.name, first_layer.name, "query"]
            )
            parameters = {"f": "json", "where": where}

            logger.debug("Getting features from %s", url)

            response = self._session.get(url, params=parameters)
            try:
                response_dict = response.json()
            except KeyError:
                return []
            try:
                features = response_dict["features"]
            except KeyError:
                logger.exception(
                    "Error running query %s. Server responded with %s",
                    url,
                    response_dict,
                )
                return []

            if complete is True:
                return response_dict
            else:
                return (
                    # the field name of the object id
                    # obviously in 99% it is "ID" but could be something else
                    response_dict["objectIdFieldName"],
                    [
                        EmsQueryResult(
                            self._session,
                            self._project,
                            self.name,
                            feature["attributes"],
                        )
                        for feature in features
                    ],
                )
        return []

    def clone_objectclass(self, objectclass_name):
        """
        Clone the objectClass

        Args:
            new objectclass name

        Returns:
            :bool: True if successfull, False otherwise
        """

        url = "/".join(["rest", self._project, self.name, "cloneObjectClass"])
        data = {
            "f": "json",
            "newObjectClassName": objectclass_name,
        }

        logger.debug(
            "Cloning the given objectclass %s to the new objectclass %s",
            self.name,
            objectclass_name,
        )

        response = self._session.post(url, params=data)
        try:
            response_dict = response.json()
        except KeyError:
            return False

        try:
            if response_dict["cloneObjectClassResult"]["success"] is False:
                error_dictionary = response_dict["cloneObjectClassResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return False
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return False

    @property
    def tables(self):
        """
        Get the tables of an objectclass

        Args:
            -

        Returns:
            if successfull
                dictionary of tables
            else
                None
        """

        url = "/".join(["rest", self._project, self.name, "tables"])
        data = {
            "f": "json",
        }

        logger.debug("Query the tables of the objectclass %s", self.name)

        response = self._session.get(url, data=data)
        try:
            response_dict = response.json()
        except KeyError:
            logger.error("Error in tables query")
            raise ObjectClassException("Error in tables query")
        try:
            table_type = response_dict[0]["type"]
            logger.debug("First table type: '%s'", table_type)
            return response_dict
        except KeyError:
            logger.error("Error in tables query")
            raise ObjectClassException("Error in tables query")

    def update_objectclass(self, objectclass_name):
        """
        Updates the objectClass

        Args:
            objectclass_name: objectclass name to update

        Returns:
            :bool: True if successfull: False otherwise
        """

        url = "/".join(["rest", self._project, self.name, "updateObjectClass"])
        data = {"f": "json"}

        logger.debug("Updating the objectclass %s.", self.name)

        response = self._session.post(url, params=data)
        try:
            response_dict = response.json()
        except KeyError:
            return False

        try:
            if response_dict["updateObjectClassResult"]["success"] is False:
                error_dictionary = response_dict["updateObjectClassResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return False
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return False

        return True

    def variants(self, feature_id):
        """
        Return the variants of the provided ID

        :param feature_id: The ID of the feature.
        :type feature_id: int
        :returns:  list -- the list of variants, None on error.
        :raises:
        """
        # TODO A lot of things to fix
        url = "/".join(["rest", self._project, self.name, str(feature_id), "variants"])
        parameters = {"f": "json"}

        logger.debug("Getting variants from %s", url)

        response = self._session.get(url, params=parameters)
        try:
            response_dict = response.json()
        except KeyError:
            return None

        logger.debug(response_dict)
        try:
            variants = response_dict["variants"]
            return variants
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return None
        else:
            return None

    def create_variant(self, feature_id, variant_id, variants):
        """Create a new catalog for the current project.

        Args:
            str: name: the name of the catalog,
                 remember: the name may not contain blanks,
                 the name will be upper case in the database
            description: the description of the catalog
            fields: catalog entries

        Returns:
            bool: True = successfull, False otherwise
        """

        url = "/".join(
            [
                "rest",
                self._project,
                self.name,
                str(feature_id),
                "variant",
                str(variant_id),
                "create",
            ]
        )
        request_parameters = {"f": "json", "variants": json.dumps(variants)}

        response_dict = self._session.post(url, params=request_parameters)

        try:
            if response_dict["addCatalogResult"]["success"] is False:
                error_dictionary = response_dict["addCatalogResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return False
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return False

        logger.info(
            "Added successfully the catalog '%s'.",
        )
        return True

    def __len__(self):
        """
        Returns the total number of features
        """
        return self._total_features

    def __repr__(self):
        """
        The name of the ObjectClass
        """
        return self.name



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import json
import logging

from anytree import Node, PreOrderIter, RenderTree, find_by_attr
from anytree.exporter import DictExporter
from anytree.render import AsciiStyle

from .emscatalog import EmsCatalog
from .emsobjectclass import ObjectClass

logger = logging.getLogger(__file__)


class EmsProjectException(Exception):
    """
    Base Exception for EmsProject
    """


class EmsProject:
    """
    Project
    """

    def __init__(self, session, project_definition):
        """
        The class will be initialized using the answer from EMS.
        With this solution we have access to all the EMS attributes
        """

        self._session = session
        self.name = ""  # because of pylint

        for key, value in project_definition.items():
            setattr(self, key, value)

    def __repr__(self):
        return str(self.name)

    @property
    def catalogs(self):
        """
        The list of all catalogs for the project
        """

        url = "/".join(["rest", self.name, "catalogs"])
        request_parameters = {"f": "json"}

        response_dict = self._ems_request(url, request_parameters)

        try:
            catalogs = response_dict["catalogs"]
        except KeyError:
            logger.debug("Getting catalogs %s", response_dict)
            raise EmsProjectException("There are no catalogs in the response")
        return [
            EmsCatalog(
                self._session,
                name=catalog["name"],
                id=catalog["id"],
                desc=catalog["description"],
                fields=catalog["fields"],
            )
            for catalog in catalogs
        ]

    def catalog(self, name):
        """
        Return the catalog.

        Args:
            string: name The name of the catalog
        """

        url = "/".join(["rest", self.name, name.upper(), "query"])
        request_parameters = {"f": "json", "query": "1=1"}

        response_dict = self._ems_request(url, request_parameters)
        catalog = EmsCatalog(
            self._session,
            name=name,
            entries=response_dict["entries"],
            fields=response_dict["fields"],
        )

        return catalog

    @property
    def objectclasses(self, count_features=False):
        """
        The list of the objectClasses for the current Project
        """

        url = "/".join(["rest", self.name, "objectClasses"])
        request_parameters = {"f": "json"}

        response_dict = self._ems_request(url, request_parameters)

        return [
            ObjectClass(
                session=self._session,
                project=self.name,
                object_definition=objectclass,
                count_features=count_features,
            )
            for objectclass in response_dict["objectClasses"]
        ]

    def objectclass(self, name):
        """
        Returns the object with the specified name.

        Examples:
            Return the objectClass with the provided name

            >>> print(my_project.objectclass("test_project"))
            test_project

        Args:
            name: Name of the objectClass

        Returns:
            objectclass: An Objectclass instance
        """

        url = "/".join(["rest", self.name, "objectClasses"])
        request_parameters = {"f": "json"}

        response_dict = self._ems_request(url, request_parameters)
        objectclass = next(
            iter(filter(lambda x: x["name"] == name, response_dict["objectClasses"])),
            None,
        )

        if objectclass is not None:
            return ObjectClass(
                session=self._session, project=self.name, object_definition=objectclass
            )

        logging.info("There is no objectClass with name = %s", name)
        return None

    def add_objectclass(self, objectclass):
        """
        Create a new objectclass for a given project.

        Args:
            objectclass: entries necessary for the objectclass (layers, tables, ..)

        Returns:
            integer: the new object id, -99: if an error occurs

        Raises:
            EmsProjectException
        """
        if self.name is None:
            raise EmsProjectException("You have to set the name of a project")
        if objectclass is None:
            raise EmsProjectException("You have to provide a valid objectclass")

        url = "/".join(["rest", self.name, "addObjectClass"])

        parameters = {}
        for key, value in objectclass.items():
            if isinstance(value, (list, dict)):
                parameters[key] = json.dumps(value)
            else:
                parameters[key] = value

        response_dict = self._ems_request(url, parameters, "POST")

        try:
            if response_dict["addObjectClassResult"]["success"] is False:
                error_dictionary = response_dict["addObjectClassResult"]["error"]
                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return -99
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return -99

        logger.info("Added successfully the objectclass '%s'.", objectclass["name"])
        return response_dict["addObjectClassResult"]["objectId"]

    def delete_objectclass(self, objectclass_id):
        """
        Deletes an objectclass for a given project.

        Args:
            project_name: the name of the project
            objectclass: the name of the objectclass

        Returns:
            bool: True = successfull, False otherwise
        """

        if isinstance(objectclass_id, int) is not True:
            logger.error("The parameter '%s' is not an integer value", objectclass_id)

        url = "/".join(["rest", self.name, "deleteObjectClass"])

        request_parameters = {"f": "json", "objectClassId": objectclass_id}
        response_dict = self._ems_request(url, request_parameters, "POST")

        if response_dict["deleteObjectClassResult"]["success"] is False:
            logger.error(
                "Message: %s",
                response_dict["deleteObjectClassResult"]["error"]["message"],
            )
            logger.error(
                "Description: %s",
                response_dict["deleteObjectClassResult"]["error"]["description"],
            )
            logger.error(
                "Code: %s", response_dict["deleteObjectClassResult"]["error"]["code"]
            )
            return False

        logger.info("Deleted the objectclass with id %s.", objectclass_id)
        return True

    def add_catalog(self, name, description=None, fields=None):
        """Create a new catalog for the current project.

        Args:
            str: name: the name of the catalog,
                 remember: the name may not contain blanks,
                 the name will be upper case in the database
            description: the description of the catalog
            fields: catalog entries

        Returns:
            bool: True = successfull, False otherwise
        """
        if fields is None:
            fields = []

        url = "/".join(["rest", self.name, "addCatalog"])
        request_parameters = {"f": "json", "name": name, "fields": json.dumps(fields)}

        if description is not None:
            request_parameters["description"] = description

        response_dict = self._ems_request(url, request_parameters, "POST")

        try:
            if response_dict["addCatalogResult"]["success"] is False:
                error_dictionary = response_dict["addCatalogResult"]["error"]

                logger.error("Message: %s", error_dictionary["message"])
                logger.error("Description: %s", error_dictionary["description"])
                logger.error("Code: %s", error_dictionary["code"])
                return False
        except KeyError:
            error_dictionary = response_dict["error"]
            logger.error("Message: %s", error_dictionary["message"])
            logger.error("Description: %s", error_dictionary["description"])
            logger.error("Code: %s", error_dictionary["code"])
            return False

        logger.info("Added successfully the catalog '%s'.", name)
        return True

    def delete_catalog(self, catalog_name):
        """
        Delete a catalog by name for the current project.

        Args:
            catalog_name: the name of the catalog

        Returns:
            bool: True = successfull, False otherwise
        """

        # firstly a mapping of the name to the name is needed
        url = "/".join(["rest", self.name, "catalogs"])
        request_parameters = {
            "f": "json",
        }
        catalog_response = self._ems_request(url, request_parameters, "GET")
        catalogs = catalog_response.get("catalogs")

        catalogs_search = list(
            filter(lambda catalog: catalog["name"] == catalog_name.upper(), catalogs)
        )

        if catalogs_search:
            catalog_id = catalogs[0]["id"]
        else:
            logger.error("There is no catalog called %s", catalog_name)
            return False

        # now we have a valid id
        url = "/".join(["rest", self.name, "deleteCatalog"])
        request_parameters = {
            "f": "json",
            "catalogId": catalog_id,
        }
        response_dict = self._ems_request(url, request_parameters, "POST")

        response_result = response_dict.get("deleteCatalogResult")
        # With a wrong id there is no key deleteCatalogResult
        if response_result is None:
            logger.debug("Response to deleteCatalog %s", response_result)
            logger.error("The provided id does not exist")
            return False

        request_success = response_result.get("success")

        if request_success:
            logger.info("Deleted successfully the catalog ")
            return True
        else:
            logger.error("Can not delete Catalog ")
            return False

    def variants_tree(self):
        """
        Return a dict and a Iterable with the traversing structure of the variants tree
        """

        complete_list = []
        search_expression = "ID>=1"

        objectclasses = self.objectclasses

        # Find the VNT Master Objectclass
        variant_master_objectclass = next(
            (
                item for item in objectclasses if item.objectClassType == "VNTMASTER"
            ),  # type: ignore
            None,
        )

        # It seems that there are no variants
        if variant_master_objectclass is None:
            logger.warning("It seems that there are no variants")
            return {}

        # Query the variant master to get all the features

        variant_master_layer = variant_master_objectclass.layers[0]
        query = variant_master_layer.query(
            where=search_expression, returnGeometry=False
        )
        # TODO Refactor
        variant_master_features = list(query.resolve())[0]["features"]

        variant_master_ids = [
            feature["attributes"]["ID"] for feature in variant_master_features
        ]

        # Query Variants in the Object using the id
        for variant_master_id in variant_master_ids:
            search_expression = "ID={}".format(variant_master_id)

            current_masters = variant_master_objectclass.query(where=search_expression)[
                1
            ]

            # We request all the variants pro master
            for current_master in current_masters:
                variant_list = current_master.variants(variant_master_id)

                # We sort all the variants. This help to build the variant because
                # the variant id will be always bigger
                sorted_variant = sorted(variant_list, key=lambda k: k["ID"])
                complete_list.extend(sorted_variant)

        # Crete the tree, this is the first node
        root_node = Node("0")

        for variant in complete_list:
            variant_full_name = "{0} (id={1})".format(variant["NAME"], variant["ID"])
            if variant["PARENT"] == -1:
                Node(
                    name=str(variant["ID"]),
                    id=variant["ID"],
                    variant_name=variant_full_name,
                    parent=find_by_attr(root_node, "0"),
                    master=True,
                    details=variant,
                )
            else:
                Node(
                    name=str(variant["ID"]),
                    id=variant["ID"],
                    variant_name=variant_full_name,
                    parent=find_by_attr(root_node, str(variant["PARENT"])),
                    master=False,
                    details=variant,
                )

        # This is the dictionary with the tree
        tree_exporter = DictExporter()
        logger.debug(tree_exporter.export(root_node))

        # This is the iterable with all the traversing structure
        # https://anytree.readthedocs.io/en/latest/api/anytree.iterators.html#anytree.iterators.preorderiter.PreOrderIter
        iterable_tree = PreOrderIter(root_node)

        logger.info(
            "Variants tree %s:",
            RenderTree(root_node, style=AsciiStyle()).by_attr("variant_name"),
        )

        # Return the 2 items
        try:
            variant_tree = tree_exporter.export(root_node)[0]["children"]
        except KeyError:
            variant_tree = tree_exporter.export(root_node)
        return variant_tree, iterable_tree

    @property
    def has_variants(self):
        # type: () -> bool

        """
        Check if the project contains variants
        """

        variant_master_objectclass = next(
            (
                item
                for item in self.objectclasses
                if item.objectClassType == "VNTMASTER"
            ),
            None,
        )

        return variant_master_objectclass is not None

    def _ems_request(self, url, parameters, method="GET"):
        """
        Return a json response from a session
        """
        logger.debug("New request to %s parameters %s", url, parameters)

        if method == "POST":
            response = self._session.post(url, data=parameters)
        else:
            response = self._session.get(url, params=parameters)

        if response.status_code != 200:
            logger.error(
                "Error talking with ems. Url: %s Parameters: %s Status code: %s",
                url,
                parameters,
                response.status_code,
            )
            raise EmsProjectException("Error talking with EMS")
        try:
            json_response = response.json()
        except ValueError as exception:
            logger.exception(
                "Response is not json. You are probably on the login website:%s",
                exception,
            )
        else:
            return json_response



#
# The MIT License (MIT)
# Copyright (c) 2022 M.O.S.S. Computer Grafik Systeme GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the Software
# is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT,TORT OR OTHERWISE, ARISING FROM, OUT
# OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

import logging

logger = logging.getLogger(__file__)


class GeojsonConvertException(Exception):
    """
    Exception for the GeojsonConvert
    """


class GeojsonConverter:
    """
    The class to convert a GeoJSON to ESRI

    """

    @staticmethod
    def _convert_geometry(geometry):
        """
        Convert the geometry into the ESRI notation

        :param geometry: The GDAL geometry dict.
        :type geometry: dict
        :returns:  dictionary -- a dictionary depending on the type of geometry, i.e.
        Point, MultiPoint, Polygon, MultiPolygon, LineString, MultiLineString.
        :raises: GeojsonConvertException
        """

        geometry_type = geometry.get("type")
        coordinates = geometry.get("coordinates")

        if geometry_type is None:
            logger.error("Error in geojson defintion: %s", geometry)
            raise GeojsonConvertException

        # Point
        if geometry_type == "Point":
            try:
                point_x = coordinates[0]
                point_y = coordinates[1]
            except IndexError:
                logger.error("Invalid geometry definition: %s", coordinates)
                raise GeojsonConvertException
            else:
                return {"x": point_x, "y": point_y}

        # MultiPoint
        if geometry_type == "MultiPoint":
            all_points = []
            for point in coordinates:
                all_points += [point]
            return {"points": all_points}

        # Polygon
        if geometry_type == "Polygon":
            # logger.info("Polygon %s", coordinates)
            return {"rings": coordinates}
        # MultiPoygon
        if geometry_type == "MultiPolygon":
            # logger.info(coordinates)
            all_polygons = []
            for polygon in coordinates:
                all_polygons += polygon
            return {"rings": all_polygons}

        # LineString
        if geometry_type == "LineString":
            # logger.debug("Converting LineString %s", coordinates)
            return {"paths": [coordinates]}

        # MultiLineString
        if geometry_type == "MultiLineString":
            # logger.debug("Converting LineString %s", coordinates)
            all_lines = []
            for line in coordinates:
                all_lines += line
            return {"paths": [all_lines]}

        raise GeojsonConvertException("Unknown geometry type: %s" % geometry_type)

    @staticmethod
    def to_esri(feature):
        """
        Converts a GDAL feature to an ESRI feature

        :param feature: The feature dict containing "properties" and "geometry".
        :type feature: dict
        :returns:  dict -- an ESRI conform feature dictionary
        :raises: GeojsonConvertException
        """
        esri_output = {}

        # Attributes

        try:
            esri_output["attributes"] = feature["properties"]
        except KeyError:
            logger.debug("There is  no 'properties' key in the current feature.")
            esri_output["attributes"] = {}

        # Geometry
        try:
            geometry = feature["geometry"]
        except KeyError:
            logger.error("There is  no 'geometry' key in the current feature.")
            raise GeojsonConvertException
        else:
            esri_output["geometry"] = GeojsonConverter._convert_geometry(geometry)

        return esri_output
